from __future__ import annotations

from contextlib import contextmanager
from dataclasses import dataclass
from datetime import datetime
from time import time
from typing import Literal, Optional, Sequence
from uuid import uuid4

import httpx
from minmodkg.config import MNO_NS, MNR_NS, SPARQL_ENDPOINT, SPARQL_UPDATE_ENDPOINT
from minmodkg.misc.utils import group_by_key
from rdflib import Graph

Triple = tuple[str, str, str]


@dataclass
class Triples:
    triples: Sequence[Triple] | set[Triple]


def sparql(query: str, endpoint: str, type: Literal["query", "update"] = "query"):
    """Run SPARQL query/update on the given endpoint. This function has been tested on Apache Jena Fuseki"""
    final_query = (
        """
    PREFIX dcterms: <http://purl.org/dc/terms/>
    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    PREFIX : <https://minmod.isi.edu/ontology/>
    PREFIX mnr: <https://minmod.isi.edu/resource/>
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    PREFIX owl: <http://www.w3.org/2002/07/owl#>
    PREFIX gkbi: <https://geokb.wikibase.cloud/entity/>
    PREFIX gkbt: <https://geokb.wikibase.cloud/prop/direct/>
    PREFIX geo: <http://www.opengis.net/ont/geosparql#>
    PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
    \n"""
        + query
    )
    # send query
    response = httpx.post(
        url=endpoint,
        data={type: final_query},
        headers={
            "Content-Type": "application/x-www-form-urlencoded",
            "Accept": "application/sparql-results+json",  # Requesting JSON format
        },
        verify=False,
        timeout=None,
    ).raise_for_status()
    return response


def sparql_construct(query: str, endpoint: str = SPARQL_ENDPOINT) -> Graph:
    resp = sparql(query, endpoint, type="query")
    g = Graph()
    g.parse(data=resp.text, format="turtle")
    return g


def sparql_delete_insert(
    delete: str | Graph | Triples | Sequence[Triples | Graph],
    insert: str | Graph | Triples | Sequence[Triples | Graph],
    endpoint: str = SPARQL_UPDATE_ENDPOINT,
):
    parts = ["DELETE {"]
    if not isinstance(delete, str):
        sub = serialize_triples(delete)
        parts.extend(sub)
    else:
        parts.append(delete)
    parts.append("\n} INSERT {")

    if not isinstance(insert, str):
        sub = serialize_triples(insert)
        parts.extend(sub)
    else:
        parts.append(insert)
    parts.append("\n} WHERE {}")
    return sparql("".join(parts), endpoint, type="update")


def sparql_delete(
    query: str | Graph | Triples | Sequence[Triples | Graph],
    endpoint: str = SPARQL_UPDATE_ENDPOINT,
):
    if not isinstance(query, str):
        query_parts = ["DELETE DATA {"]
        query_parts.extend(serialize_triples(query))
        query_parts.append("\n}")
        query = "".join(query_parts)

    return sparql(query, endpoint, type="update")


def sparql_insert(
    query: str | Graph | Triples | Sequence[Triples | Graph],
    endpoint: str = SPARQL_UPDATE_ENDPOINT,
):
    if not isinstance(query, str):
        query_parts = ["INSERT DATA {"]
        query_parts.extend(serialize_triples(query))
        query_parts.append("\n}")
        query = "".join(query_parts)

    return sparql(query, endpoint, type="update")


def sparql_query(
    query: str,
    endpoint=SPARQL_ENDPOINT,
    keys: Optional[list[str]] = None,
) -> list[dict]:
    response = sparql(query, endpoint, type="query")

    if response.status_code != 200:
        raise Exception(response.text)

    qres = response.json()["results"]["bindings"]

    if len(qres) == 0:
        return []

    sample = {}
    for res in qres:
        for key in res:
            if key not in sample:
                sample[key] = res[key]
    if keys is not None:
        for key in keys:
            if key not in sample:
                sample[key] = {"type": "bnode"}

    output = [{} for _ in range(len(qres))]
    for key, val in sample.items():
        if val["type"] in {"uri", "bnode"}:
            for i, row in enumerate(qres):
                if key in row:
                    output[i][key] = row[key]["value"]
                else:
                    output[i][key] = None
        else:
            assert val["type"] == "literal", val["type"]
            if val.get("datatype") in {
                None,
                "http://www.opengis.net/ont/geosparql#wktLiteral",
            }:
                for i, row in enumerate(qres):
                    if key in row:
                        output[i][key] = row[key]["value"]
                    else:
                        output[i][key] = None
            elif val["datatype"] in {
                "http://www.w3.org/2001/XMLSchema#integer",
            }:
                for i, row in enumerate(qres):
                    if key in row:
                        try:
                            output[i][key] = int(row[key]["value"])
                        except ValueError:
                            # we think it's an integer but it's not, so we don't attempt to parse this column
                            for i, row in enumerate(qres):
                                if key in row:
                                    output[i][key] = row[key]["value"]
                                else:
                                    output[i][key] = None
                            break
                    else:
                        output[i][key] = None
            elif val["datatype"] in {
                "http://www.w3.org/2001/XMLSchema#decimal",
                "http://www.w3.org/2001/XMLSchema#double",
                "http://www.w3.org/2001/XMLSchema#float",
            }:
                for i, row in enumerate(qres):
                    if key in row:
                        output[i][key] = float(row[key]["value"])
                    else:
                        output[i][key] = None
            elif val["datatype"] == "http://www.w3.org/2001/XMLSchema#date":
                for i, row in enumerate(qres):
                    if key in row:
                        rowval = row[key]["value"]
                        try:
                            output[i][key] = datetime.strptime(rowval, "%Y-%m-%d")
                        except ValueError:
                            output[i][key] = datetime.strptime(rowval, "%Y-%m")
                    else:
                        output[i][key] = None
            elif val["datatype"] == "http://www.w3.org/2001/XMLSchema#boolean":
                for i, row in enumerate(qres):
                    if key in row:
                        output[i][key] = row[key]["value"] == "true"
                    else:
                        output[i][key] = None
            else:
                raise NotImplementedError(val)
    return output


def serialize_triples(query: Graph | Triples | Sequence[Triples | Graph]) -> list[str]:
    parts = []
    if isinstance(query, Graph):
        for s, p, o in query:
            parts.append(f"\n{s.n3()} {p.n3()} {o.n3()} .")
    elif isinstance(query, Triples):
        for s, p, o in query.triples:
            parts.append(f"\n{s} {p} {o} .")
    elif not isinstance(query, str):
        assert isinstance(query, Sequence)
        for g in query:
            if isinstance(g, Graph):
                for s, p, o in g:
                    parts.append(f"\n{s.n3()} {p.n3()} {o.n3()} .")
            else:
                assert isinstance(g, Triples)
                for s, p, o in g.triples:
                    parts.append(f"\n{s} {p} {o} .")
    return parts


def has_uri(uri: str, endpoint: str = SPARQL_ENDPOINT) -> bool:
    query = (
        """
    SELECT ?uri WHERE { 
        ?uri ?p ?o 
        VALUES ?uri { <%s> }
    }
    LIMIT 1"""
        % uri
    )
    qres = sparql_query(query, endpoint)
    return len(qres) > 0


class Transaction:
    """Steps to perform a transaction:

    t10 -> insert the lock
    t20 -> check if there is another lock.
    t30 -> perform the query/update (update query can only be done once -- because no rollback)
    t40 -> release the lcok
    """

    def __init__(
        self,
        objects: list[str],
        query_endpoint: str = SPARQL_ENDPOINT,
        update_endpoint: str = SPARQL_UPDATE_ENDPOINT,
        timeout_sec: float = 300,
    ):
        self.objects = []
        value_filter_parts = []

        for obj in objects:
            if obj.startswith("http://") or obj.startswith("https://"):
                self.objects.append(obj)
                value_filter_parts.append(f"<{obj}>")
            elif obj.startswith("mnr:"):
                self.objects.append(MNR_NS + obj[4:])
                value_filter_parts.append(obj)
            elif obj.startswith("mno:"):
                self.objects.append(MNO_NS + obj[4:])
                value_filter_parts.append(obj)
            else:
                raise ValueError("Invalid object " + obj)

        self.value_query = " ".join(value_filter_parts)
        self.query_endpoint = query_endpoint
        self.update_endpoint = update_endpoint
        self.timeout_sec = timeout_sec
        self.lock = None

    @contextmanager
    def transaction(self):
        self.insert_lock()
        if not self.does_lock_success():
            raise Exception(
                "The objects are being edited by another one. Please try again later."
            )
        # yield so the caller can perform the transaction
        yield
        self.remove_lock()

    def insert_lock(self):
        assert self.lock is None
        self.lock = f"{str(uuid4())}::{time() + self.timeout_sec}"
        sparql_insert(
            Triples([(f"<{obj}>", ":lock", f'"{self.lock}"') for obj in self.objects]),
            endpoint=self.update_endpoint,
        )

    def does_lock_success(self):
        lst = sparql_query(
            """
    SELECT ?source ?lock
    WHERE {
        ?source :lock ?lock 
        VALUES ?source { %s }
    }"""
            % self.value_query,
            keys=["source", "lock"],
            endpoint=self.query_endpoint,
        )

        obj2locks: dict[str, list[str]] = group_by_key(lst, key="source", value="lock")
        if len(obj2locks) != len(self.objects):
            return False

        now = time()
        for obj in self.objects:
            if obj not in obj2locks:
                return False
            locks = [
                lock
                for lock in obj2locks[obj]
                if lock == self.lock or float(lock.split("::")[1]) >= now
            ]
            if len(locks) != 1 or locks[0] != self.lock:
                return False

        return True

    def remove_lock(self):
        sparql_delete(
            Triples([(f"<{obj}>", ":lock", f'"{self.lock}"') for obj in self.objects]),
            endpoint=self.update_endpoint,
        )



from __future__ import annotations

from collections import defaultdict

from minmodkg.schema.make_diagram import (
    get_class_in_order,
    get_domain_to_props,
    hasSubPropertyOf,
    load_ontology,
    mno,
    read_a_list,
    schema_dir,
)
from rdflib import OWL, RDF, RDFS, SH, BNode, Graph, Namespace, URIRef

mnos = Namespace("https://minmod.isi.edu/ontology-shape/")


def make_shapes(skip_parent_class: bool = True):
    """We do not"""
    g = load_ontology()
    sg = Graph()
    sg.bind("sh", SH)
    sg.bind("mnos", mnos)

    for prefix, ns in g.namespaces():
        sg.bind(prefix, ns)

    props, domain2props = get_domain_to_props(g)

    for subj in get_class_in_order(g, props, domain2props):
        clsname = subj[len(mno) :]
        prop2restriction = defaultdict(list)
        for obj in g.objects(subj, RDFS.subClassOf):
            if (obj, RDF.type, OWL.Restriction) not in g:
                continue

            (prop,) = list(g.objects(obj, OWL.onProperty))
            prop2restriction[prop].append(obj)

        prop_targets = []
        for prop in domain2props.get(subj, []):
            assert isinstance(prop, URIRef)
            if (prop, hasSubPropertyOf, None) in g:
                continue

            targets = props[prop]["range"]
            for restriction in prop2restriction[prop]:
                if (restriction, OWL.allValuesFrom, None) in g:
                    (val,) = list(g.objects(restriction, OWL.allValuesFrom))
                    if isinstance(val, URIRef):
                        vals = [val]
                    else:
                        vals = read_a_list(g, val)
                    targets = [type for type in targets if type in vals]
                    assert len(targets) > 0

            targets = [t for t in targets if t in mno]
            if len(targets) == 0:
                continue

            prop_targets.append(
                (
                    prop,
                    targets,
                )
            )

        if len(prop_targets) == 0:
            # has no target links, we skip it as drepr validate for literal values
            continue

        if skip_parent_class and len(list(g.subjects(RDFS.subClassOf, subj))) > 0:
            # skip parent classes -- only plot the leaf classes
            continue

        subj_shp = mnos[clsname + "Shape"]
        sg.add((subj_shp, RDF.type, SH.NodeShape))
        sg.add((subj_shp, SH.targetClass, subj))

        for prop, targets in prop_targets:
            constraint = BNode()
            sg.add((subj_shp, SH.property, constraint))
            sg.add((constraint, SH.path, prop))
            for target in targets:
                sg.add((constraint, SH["class"], target))
            sg.add((constraint, SH.severity, SH.Violation))

    sg.serialize(schema_dir / "shapes.ttl", format="ttl")


if __name__ == "__main__":
    make_shapes()



from __future__ import annotations

from datetime import datetime
from enum import Enum
from hashlib import sha256
from typing import Literal, Optional
from uuid import uuid4

from passlib.context import CryptContext
from sqlmodel import Field, SQLModel, String

pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")


class UserBase(SQLModel):
    username: str = Field(max_length=100, primary_key=True)
    name: str
    email: str

    def is_system(self):
        return self.username in ("inferlink", "sri", "umn", "usc")


class User(UserBase, table=True):
    salt: str = Field(default_factory=lambda: str(uuid4()))
    password: str = Field(max_length=64)
    scope: Literal["admin", "user"] = Field(default="user", sa_type=String)

    def encrypt_password(self):
        self.password = pwd_context.hash(self.salt + self.password)

    def verify_password(self, password: str):
        return pwd_context.verify(self.salt + password, self.password)


class UserPublic(UserBase): ...


class UserCreate(UserBase):
    password: str


class UserUpdate(UserBase):
    name: Optional[str]
    email: Optional[str]
    password: Optional[str]



from __future__ import annotations

import os
from contextlib import asynccontextmanager

from fastapi import FastAPI
from minmodkg.api.internal import admin
from minmodkg.api.models.db import create_db_and_tables
from minmodkg.api.routers import (
    commodity,
    dedup_mineral_site,
    deposit_type,
    lod,
    login,
    mineral_site,
    stats,
    unit,
)
from minmodkg.config import API_PREFIX, LOD_PREFIX


@asynccontextmanager
async def lifespan(app: FastAPI):
    create_db_and_tables()
    yield


app = FastAPI(
    openapi_url=f"{API_PREFIX}/openapi.json",
    docs_url=f"{API_PREFIX}/docs",
    lifespan=lifespan,
)

app.include_router(commodity.router, prefix=API_PREFIX)
app.include_router(dedup_mineral_site.router, prefix=API_PREFIX)
app.include_router(deposit_type.router, prefix=API_PREFIX)
app.include_router(lod.router, prefix=LOD_PREFIX)
app.include_router(login.router, prefix=API_PREFIX)
app.include_router(mineral_site.router, prefix=API_PREFIX)
app.include_router(stats.router, prefix=API_PREFIX)
app.include_router(unit.router, prefix=API_PREFIX)
app.include_router(admin.router, prefix=f"{API_PREFIX}/admin")



from __future__ import annotations

import httpx
import orjson
import shapely.wkt
from minmodkg.grade_tonnage_model import Mt_unit, percent_unit
from minmodkg.integrations.cdr_helper import (
    MINMOD_API,
    MINMOD_SYSTEM,
    CDRHelper,
    MinmodHelper,
)
from minmodkg.integrations.cdr_schemas import (
    DedupSite,
    DedupSiteRecord,
    DepositType,
    DepositTypeCandidate,
)
from tqdm import tqdm
from minmodkg.config import MNR_NS

def replace_deposit_types():
    deposit_type_resp = httpx.get(
        f"{MINMOD_API}/deposit_types",
        verify=False,
        timeout=None,
    )
    deposit_type_resp.raise_for_status()

    deposit_types = []
    for record in deposit_type_resp.json():
        deposit_types.append(
            orjson.loads(
                DepositType(
                    id=record["uri"][len(MNR_NS) :],
                    name=record["name"],
                    group=record["group"],
                    environment=record["environment"],
                ).model_dump_json(exclude_none=True)
            )
        )

    CDRHelper.truncate(CDRHelper.DepositType)
    CDRHelper.upload_collection(CDRHelper.DepositType, deposit_types)


def upload_ta2_output(
    commodity: str,
    norm_tonnage_unit: str,
    norm_grade_unit: str,
    no_upload: bool = False,
):
    resp = httpx.get(
        f"{MINMOD_API}/dedup_mineral_sites/{commodity}",
        params={
            "norm_tonnage_unit": norm_tonnage_unit,
            "norm_grade_unit": norm_grade_unit,
        },
        verify=False,
        timeout=None,
    )
    resp.raise_for_status()

    if no_upload:
        return

    dedup_sites = resp.json()

    dpt2id = MinmodHelper.get_deposit_type_to_id()
    commodity_id2name = MinmodHelper.get_commodity_id2name()
    unit_uri2name = MinmodHelper.get_unit_uri2name()

    inputs = []
    for group in dedup_sites:
        if (
            group["best_loc_wkt"] is not None
            and group["best_loc_wkt"] != "MULTIPOINT()"
        ):
            assert group["best_loc_wkt"] != ""
            assert group["best_loc_centroid_epsg_4326"] is not None, (
                "Invalid WKT",
                group["best_loc_wkt"],
            )
            centroid = group["best_loc_centroid_epsg_4326"]
        else:
            centroid = ""

        inputs.append(
            orjson.loads(
                DedupSite(
                    sites=[
                        DedupSiteRecord(
                            mineral_site_id=site["ms"],
                            name=site["ms_name"] or "",
                            country=site["country"] or "",
                            province=site["state_or_province"] or "",
                            site_rank=site["ms_rank"] or "",
                            site_type=site["ms_type"] or "",
                        )
                        for site in group["sites"]
                    ],
                    commodity=commodity_id2name[group["commodity"]],
                    contained_metal=group["total_contained_metal"],
                    contained_metal_units=unit_uri2name[
                        group["total_contained_metal_unit"]
                    ],
                    tonnage=group["total_tonnage"],
                    tonnage_units=unit_uri2name[group["total_tonnage_unit"]],
                    grade=group["total_grade"],
                    grade_units=unit_uri2name[group["total_grade_unit"]],
                    crs=group["best_loc_crs"] or "",
                    centroid=centroid,
                    geom=group["best_loc_wkt"],
                    deposit_type_candidate=[
                        DepositTypeCandidate(
                            deposit_type_id=dpt2id[dt["name"]],
                            confidence=dt["confidence"],
                            source=dt["source"],
                        )
                        for dt in group["deposit_types"]
                    ],
                    system=MINMOD_SYSTEM,
                    system_version="1.1.3",
                    data_snapshot=commodity,
                    data_snapshot_date="",
                ).model_dump_json(exclude_none=True)
            )
        )

    CDRHelper.upload_collection(CDRHelper.DedupSites, inputs)


def get_critical_commodities():
    resp = httpx.get(
        f"{MINMOD_API}/commodities?is_critical=1",
        verify=False,
        timeout=None,
    )
    resp.raise_for_status()
    return [r["name"] for r in resp.json()]


if __name__ == "__main__":
    # replace_deposit_types()
    # CDRHelper.truncate(CDRHelper.DedupSites)
    commodities = sorted(get_critical_commodities())
    commodities += ["Mica"]

    print(commodities)

    for commodity in tqdm(commodities):
        upload_ta2_output(
            commodity,
            norm_tonnage_unit=Mt_unit,
            norm_grade_unit=percent_unit,
            no_upload=True,
        )

    CDRHelper.truncate(CDRHelper.DedupSites)

    for commodity in tqdm(commodities):
        upload_ta2_output(
            commodity,
            norm_tonnage_unit=Mt_unit,
            norm_grade_unit=percent_unit,
            no_upload=False,
        )



from __future__ import annotations

from functools import lru_cache
from typing import Optional

from fastapi import APIRouter
from minmodkg.api.dependencies import SPARQL_ENDPOINT, get_snapshot_id
from minmodkg.misc import sparql_query

router = APIRouter(tags=["deposit_types"])


@router.get("/commodities")
def commodities(is_critical: Optional[bool] = None):
    commodities = get_commodities(get_snapshot_id())
    if is_critical is not None:
        if is_critical:
            commodities = [comm for comm in commodities if comm["is_critical"]]
        else:
            commodities = [comm for comm in commodities if not comm["is_critical"]]
    return commodities


@lru_cache(maxsize=1)
def get_commodities(snapshot_id: str, endpoint: str = SPARQL_ENDPOINT):
    query = """
    SELECT ?uri ?name ?is_critical
    WHERE {
        ?uri a :Commodity ;
            rdfs:label ?name ;
            :is_critical_commodity ?is_critical
    }
    """
    qres = sparql_query(query, endpoint)
    return qres



from __future__ import annotations

from typing import Annotated, Literal

import htbuilder as H
import rdflib
import rdflib.term
from fastapi import APIRouter, HTTPException, Query
from fastapi.responses import HTMLResponse
from minmodkg.api.dependencies import SPARQL_ENDPOINT
from minmodkg.config import MNO_NS, MNR_NS
from minmodkg.misc import sparql_construct
from minmodkg.misc.sparql import has_uri
from rdflib import RDF, RDFS, BNode, Graph
from rdflib import Literal as RDFLiteral
from rdflib import URIRef

router = APIRouter(tags=["lod"])


@router.get("/resource/{resource_id}")
def get_resource(
    resource_id: str, format: Annotated[Literal["html", "json"], Query()] = "html"
):
    uri = URIRef(MNR_NS + resource_id)
    if not has_uri(uri):
        raise HTTPException(status_code=404, detail="Resource not found")

    if format == "html":
        return render_entity_html(uri, SPARQL_ENDPOINT)
    if format == "json":
        return render_entity_json(uri, SPARQL_ENDPOINT)
    raise HTTPException(status_code=400, detail="Invalid format")


@router.get("/ontology/{resource_id}")
def get_ontology(
    resource_id: str, format: Annotated[Literal["html", "json"], Query()] = "html"
):  # -> HTMLResponse | dict[Any, Any]:
    uri = URIRef(MNO_NS + resource_id)
    if not has_uri(uri):
        raise HTTPException(status_code=404, detail="Resource not found")

    if format == "html":
        return render_entity_html(uri, SPARQL_ENDPOINT)
    if format == "json":
        return render_entity_json(uri, SPARQL_ENDPOINT)
    raise HTTPException(status_code=400, detail="Invalid format")


def get_entity_data(subj: URIRef, endpoint: str) -> Graph:
    # https://stackoverflow.com/questions/37186530/how-do-i-construct-get-the-whole-sub-graph-from-a-given-resource-in-rdf-graph/37213209#37213209
    # adapt from this answer to fit our need
    # we do not blindly follow all paths (<>|!<>)* and filter out the URI nodes because it follows
    # the sameAs path, which leads to explosive results
    return sparql_construct(
        """
    CONSTRUCT { 
        ?a ?b ?c . 
        ?s ?p ?o . 
        ?c rdfs:label ?cname .
        ?o rdfs:label ?oname .
        ?p rdfs:label ?pname .
        ?b rdfs:label ?bname .
    }
    WHERE {
        ?a ?b ?c .
        OPTIONAL { ?c rdfs:label ?cname . }
        OPTIONAL { ?b rdfs:label ?bname . }
        OPTIONAL { 
            ?a (!(owl:sameAs|rdf:type))+ ?s . 
            ?s ?p ?o .
            OPTIONAL { ?o rdfs:label ?oname . }
            OPTIONAL { ?p rdfs:label ?pname .}
        }
        VALUES ?a { <%s> } 
    }
"""
        % subj,
        endpoint=endpoint,
    )


def render_entity_json(subj: URIRef, endpoint: str):
    g = get_entity_data(subj, endpoint)

    def label(obj: rdflib.term.Node):
        if obj == RDFS.label:
            return "@label"
        if obj == RDF.type:
            return "@type"
        assert isinstance(obj, (URIRef, RDFLiteral))
        return obj.n3(g.namespace_manager).rsplit(":", 1)[-1]

    def make_tree(obj: rdflib.term.Node, visited: set):
        if isinstance(obj, RDFLiteral):
            return obj
        if isinstance(obj, URIRef):
            out: dict = {"@id": obj}
        else:
            assert isinstance(obj, BNode)
            out: dict = {}

        if obj in visited:
            # skip visited nodes
            if (obj, RDFS.label, None) in g:
                out["@label"] = next(g.objects(obj, RDFS.label))
            return out

        visited.add(obj)
        for p in g.predicates(obj, unique=True):
            plabel = label(p)
            subobjs = list(g.objects(obj, p))
            fmtsubobjs = [make_tree(subobj, visited) for subobj in subobjs]
            if len(fmtsubobjs) == 1:
                out[plabel] = fmtsubobjs[0]
            else:
                out[plabel] = fmtsubobjs

        return out

    out: dict = {"@id": subj}
    visited = {subj}
    for p in g.predicates(subj, unique=True):
        plabel = label(p)
        objs = list(g.objects(subj, p))
        fmtobjs = [make_tree(obj, visited) for obj in objs]
        if len(fmtobjs) == 1:
            out[plabel] = fmtobjs[0]
        else:
            out[plabel] = fmtobjs
    return out


def render_entity_html(subj: URIRef, endpoint: str):
    g = get_entity_data(subj, endpoint)

    def label(g, subj: rdflib.term.Node):
        if (subj, RDFS.label, None) in g:
            return next(g.objects(subj, RDFS.label))
        assert isinstance(subj, (URIRef, RDFLiteral))
        return subj.n3(g.namespace_manager)

    def make_tree(g, subj: rdflib.term.Node, visited: set):
        if isinstance(subj, RDFLiteral):
            return H.p(subj)
        if isinstance(subj, URIRef):
            subj_name = subj.n3(g.namespace_manager)
            if (subj, RDFS.label, None) in g:
                subj_name = next(g.objects(subj, RDFS.label))

            return H.a(href=subj)(subj_name)

        if subj in visited:
            return H.p(style="font-style: italic")("skiped as visited before")

        visited.add(subj)
        assert isinstance(subj, BNode)
        children = []
        for p, o in g.predicate_objects(subj):
            if p != RDFS.label:
                children.append((H.a(href=p)(label(g, p)), make_tree(g, o, visited)))

        return (
            H.table(_class="table")(
                *[
                    H.tr(
                        H.td(p),
                        H.td(o),
                    )
                    for p, o in children
                ]
            ),
        )

    subj_label = label(g, subj)

    children = []
    for p, o in g.predicate_objects(subj):
        if p != RDFS.label:
            children.append((H.a(href=p)(label(g, p)), make_tree(g, o, set())))

    tree = H.div(_class="container-fluid")(
        H.div(_class="row", style="margin-top: 20px; margin-bottom: 20px")(
            H.div(_class="col")(
                H.h4(
                    H.a(href=subj)(subj_label),
                ),
                H.small(_class="text-muted fw-semibold")(subj),
            )
        ),
        H.table(_class="table table-striped")(
            *[
                H.tr(
                    H.td(p),
                    H.td(o),
                )
                for p, o in children
            ]
        ),
    )
    return HTMLResponse(
        content=f"""
<html>
    <head>
        <title>{subj_label}</title>
        <link href="https://cdn.jsdelivr.net/npm/bootstrap@5.3.3/dist/css/bootstrap.min.css" rel="stylesheet" integrity="sha384-QWTKZyjpPEjISv5WaRU9OFeRpok6YctnYmDr5pNlyT2bRjXh0JMhjY6hW+ALEwIH" crossorigin="anonymous">
        <style>a {{ text-decoration: none; }}</style>
    </head>
    <body>
        {tree}
        <script src="https://cdn.jsdelivr.net/npm/@popperjs/core@2.11.8/dist/umd/popper.min.js" integrity="sha384-I7E8VVD/ismYTF4hNIPjVp/Zjvgyol6VFvRkX/vR+Vc4jQkC+hVqc2pM8ODewa9r" crossorigin="anonymous"></script>
        <script src="https://cdn.jsdelivr.net/npm/bootstrap@5.3.3/dist/js/bootstrap.min.js" integrity="sha384-0pUGZvbkm6XF6gxjEnlmuGrJXVbNuzT9qBBavbLwCsOGabYfZo0T0to5eqruptLy" crossorigin="anonymous"></script>
    </body>
</html>
                        """,
        status_code=200,
    )



import re
from collections import defaultdict
from dataclasses import asdict, dataclass, make_dataclass
from datetime import datetime
from graphlib import TopologicalSorter
from itertools import chain
from pathlib import Path
from typing import List, Optional, TypeAlias, Union

import erdantic as erd
from pydantic import BaseModel, ConfigDict, Field, create_model
from rdflib import OWL, RDF, RDFS, XSD, BNode, Graph, Literal, Namespace, URIRef

mno = Namespace("https://minmod.isi.edu/ontology/")
geo = Namespace("http://www.opengis.net/ont/geosparql#")
hasSubPropertyOf = mno.hasSubPropertyOf
schema_dir = Path(__file__).parent.parent.parent / "schema"


class URI(str): ...


class NotEmptyList(list): ...


def get_all_values(g: Graph, s: URIRef, p: URIRef, subItemOf: URIRef):
    output = list(g.objects(s, p, unique=True))
    for parent in g.objects(s, subItemOf):
        assert isinstance(parent, URIRef)
        output.extend(get_all_values(g, parent, p, subItemOf))
    return output


def load_ontology():
    ontology_file = schema_dir / "ontology.ttl"
    g = Graph()
    g.parse(ontology_file, format="turtle")

    # run basic inference to inherit domain and range
    all_props = list(g.subjects(RDF.type, OWL.ObjectProperty, unique=True)) + list(
        g.subjects(RDF.type, OWL.DatatypeProperty, unique=True)
    )
    for prop in all_props:
        assert isinstance(prop, URIRef)
        for attr in [RDFS.domain, RDFS.range]:
            existing_vals = set(g.objects(prop, attr, unique=True))
            for val in get_all_values(g, prop, attr, RDFS.subPropertyOf):
                if val not in existing_vals:
                    g.add((prop, attr, val))

    # run basic inference to inherit subClassOf
    for subj, obj in list(g.subject_objects(RDFS.subClassOf, unique=True)):
        if not isinstance(subj, URIRef) and not isinstance(obj, URIRef):
            continue

        for obj_parent in g.objects(obj, RDFS.subClassOf, unique=True):
            g.add((subj, RDFS.subClassOf, obj_parent))

    # add hasSubPropertyOf
    for s, p, o in list(g.triples((None, RDFS.subPropertyOf, None))):
        g.add((o, hasSubPropertyOf, s))

    return g


def read_a_list(g: Graph, s):
    out: list[URIRef] = []
    for pred, obj in g.predicate_objects(s):
        if pred == RDF.rest:
            out.extend(read_a_list(g, obj))
        else:
            assert isinstance(obj, URIRef)
            out.append(obj)
    return out


def get_domain_or_range(g: Graph, prop: URIRef, attr: URIRef):
    output = set()
    for obj in g.objects(prop, attr, unique=True):
        if isinstance(obj, URIRef):
            output.add(obj)
        else:
            assert (obj, OWL.unionOf, None) in g
            for sobj in read_a_list(g, next(g.objects(obj, OWL.unionOf))):
                output.add(sobj)
    return sorted(output)


def get_domain_to_props(g: Graph):
    # get list of properties
    props = {}
    domain2props = {}
    for prop in chain(
        g.subjects(RDF.type, OWL.ObjectProperty),
        g.subjects(RDF.type, OWL.DatatypeProperty),
        g.subjects(RDF.type, RDF.Property),
    ):
        assert isinstance(prop, URIRef)
        domains = get_domain_or_range(g, prop, RDFS.domain)
        ranges = get_domain_or_range(g, prop, RDFS.range)

        props[prop] = {
            "domain": domains,
            "range": ranges,
        }
        for domain in domains:
            domain2props.setdefault(domain, []).append(prop)
    domain2props = {k: sorted(v) for k, v in domain2props.items()}
    for subj in g.subjects(RDF.type, OWL.Class):
        if isinstance(subj, URIRef) and subj not in domain2props:
            domain2props[subj] = []

    for subj in domain2props:
        for parent_subj in g.objects(subj, RDFS.subClassOf, unique=True):
            if not isinstance(parent_subj, URIRef):
                continue
            for prop in domain2props.get(parent_subj, []):
                if prop not in domain2props[subj]:
                    domain2props[subj].append(prop)
    return props, domain2props


def get_data_types(g: Graph):
    namespaces = list(g.namespaces())
    datatypes = {}
    for subj in g.subjects(RDF.type, RDFS.Datatype):
        ((prefix, ns),) = [
            (prefix, ns) for prefix, ns in namespaces if str(subj).startswith(ns)
        ]
        typename = prefix + ":" + str(subj)[len(ns) :]
        datatypes[subj] = type(typename, (object,), {})
    return datatypes


def get_class_in_order(g: Graph, props, domain2props):
    # get class orders
    edges: dict[URIRef, list[URIRef]] = {
        subj: [] for subj in g.subjects(RDF.type, OWL.Class) if isinstance(subj, URIRef)
    }
    for subj in edges.keys():
        for prop in domain2props.get(subj, []):
            edges[subj].extend((t for t in props[prop]["range"] if t in edges))
    ts = TopologicalSorter(edges)
    return list(ts.static_order())


def make_er_diagram(skip_empty_field: bool = True, skip_parent_class: bool = True):
    g = load_ontology()

    # get list of properties
    props, domain2props = get_domain_to_props(g)

    # get data types
    datatypes = get_data_types(g)

    # iterate over classes and create models
    models = {}
    for subj in get_class_in_order(g, props, domain2props):
        # gather all restrictions -- not parse yet
        prop2restriction = defaultdict(list)
        for obj in g.objects(subj, RDFS.subClassOf):
            if (obj, RDF.type, OWL.Restriction) not in g:
                continue

            (prop,) = list(g.objects(obj, OWL.onProperty))
            prop2restriction[prop].append(obj)

        fields = []

        # things with label
        if (subj, RDFS.subClassOf, mno.ThingHasLabel) in g:
            fields.append(("name", str))

        if (subj, RDFS.subClassOf, mno.ThingMayHaveAltLabel) in g:
            fields.append(("aliases", Optional[list[str]]))

        if (subj, RDFS.subClassOf, mno.ThingMayHaveComment) in g:
            fields.append(("comment", Optional[str]))

        clsname = subj[len(mno) :]
        for prop in domain2props.get(subj, []):
            assert isinstance(prop, URIRef)
            if (prop, hasSubPropertyOf, None) in g:
                continue
            propname = prop[len(mno) :]
            if sum(1 for _ in g.objects(prop, RDFS.range, unique=True)) > 1:
                raise Exception("Cannot represent intersection types in pydantic")
            fieldtype = get_field_type(
                g, models, datatypes, props[prop]["range"], prop2restriction[prop]
            )
            fields.append((propname, fieldtype))

        if skip_empty_field and len(fields) == 0:
            continue

        if skip_parent_class and len(list(g.subjects(RDFS.subClassOf, subj))) > 0:
            # skip parent classes -- only plot the leaf classes
            continue

        models[clsname] = create_model(
            clsname,
            **{name: (type, Field()) for name, type in fields},  # type: ignore
            __config__=ConfigDict(arbitrary_types_allowed=True)
        )  # type: ignore

    graph = erd.create(*list(models.values()))
    graph.draw(out=schema_dir / "er_diagram.png")
    return


def get_field_type(
    g: Graph,
    models: dict[str, type],
    datatypes: dict[str, type],
    types: list[URIRef],
    restrictions: list[URIRef | BNode],
):
    if len(restrictions) > 0:
        for restriction in restrictions:
            if (restriction, OWL.allValuesFrom, None) in g:
                (val,) = list(g.objects(restriction, OWL.allValuesFrom))
                if isinstance(val, URIRef):
                    vals = [val]
                else:
                    vals = read_a_list(g, val)
                types = [type for type in types if type in vals]
                assert len(types) > 0

    norm_types = []
    for obj in types:
        if obj == XSD.string:
            norm_types.append(str)
        elif obj == XSD.integer:
            norm_types.append(int)
        elif obj == XSD.decimal:
            norm_types.append(float)
        elif obj == XSD.anyURI:
            norm_types.append(URI)
        elif obj == XSD.dateTime:
            norm_types.append(datetime)
        elif isinstance(obj, URIRef) and obj in mno:
            norm_types.append(models[str(obj)[len(mno) :]])
        elif obj in datatypes:
            norm_types.append(datatypes[obj])
        else:
            raise NotImplementedError((types, obj))

    if len(norm_types) == 1:
        norm_type = norm_types[0]
    else:
        assert len(norm_types) > 1, (types, norm_types)
        norm_type = Union[tuple(norm_types)]  # type: ignore

    if restrictions is not None:
        update = False
        for restriction in restrictions:
            for cons, cons_val in g.predicate_objects(restriction):
                if cons in (RDF.type, OWL.onProperty, OWL.allValuesFrom, OWL.onClass):
                    continue
                if cons in (OWL.cardinality, OWL.qualifiedCardinality):
                    # do nothing
                    assert isinstance(cons_val, Literal)
                    assert cons_val.value == 1
                    update = True
                elif cons == OWL.maxCardinality:
                    assert isinstance(cons_val, Literal)
                    assert cons_val.value == 1
                    norm_type = Optional[norm_type]  # type: ignore
                    update = True
                elif cons == OWL.minCardinality:
                    assert isinstance(cons_val, Literal)
                    assert cons_val.value == 1
                    norm_type = NotEmptyList[norm_type]  # type: ignore
                    update = True
                else:
                    raise NotImplementedError(cons)
        if not update:
            norm_type = list[norm_type]
    else:
        # by default, it's a list of values
        norm_type = list[norm_type]
    return norm_type


if __name__ == "__main__":
    make_er_diagram()



from __future__ import annotations

from typing import Any, Optional, TypeVar

V = TypeVar("V")


def batch(size: int, *vars, return_tuple: bool = False):
    """Batch the variables into batches of size. When vars is a single variable,
    it will return a list of batched values instead of list of tuple of batched values.

    If we want to batch a single variable to a list of tuple of batched values, set
    return_tuple to True.
    """
    output = []
    n = len(vars[0])
    if len(vars) == 1 and not return_tuple:
        for i in range(0, n, size):
            output.append(vars[0][i : i + size])
    else:
        for i in range(0, n, size):
            output.append(tuple(var[i : i + size] for var in vars))
    return output


def group_by_key(
    output: list[dict], key: str, value: Optional[str] = None
) -> dict[str, list]:
    groups = {}
    if value is None:
        for row in output:
            val = row[key]
            if val not in groups:
                groups[val] = []
            groups[val].append(row)
    else:
        for row in output:
            val = row[key]
            if val not in groups:
                groups[val] = []
            groups[val].append(row[value])
    return groups


def group_by_attr(output: list[V], attr: str) -> dict[str, list[V]]:
    groups = {}
    for row in output:
        val = getattr(row, attr)
        if val not in groups:
            groups[val] = []
        groups[val].append(row)
    return groups


def assert_isinstance(x: Any, cls: type[V]) -> V:
    assert isinstance(x, cls), x
    return x



from __future__ import annotations

import os
from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path
from typing import Optional

import httpx
import timer
from joblib import Parallel, delayed
from minmodkg.config import MNR_NS
from tqdm import tqdm

if "CDR_AUTH_TOKEN" in os.environ:
    AUTH_TOKEN = os.environ.get("CDR_AUTH_TOKEN")
else:
    keyfile = Path(os.path.expanduser("~/.config/cdr.key"))
    assert keyfile.exists()
    AUTH_TOKEN = keyfile.read_text().strip()

MINMOD_API = os.environ.get("MINMOD_API", "https://minmod.isi.edu/api/v1")
MINMOD_SYSTEM = os.environ.get("MINMOD_SYSTEM", "minmod")
CDR_API = "https://api.cdr.land/v1"

cdr_headers = {"Authorization": f"Bearer {AUTH_TOKEN}"}


@dataclass
class Endpoint:
    item: str
    collection: str
    bulk_upload: Optional[str] = None
    bulk_delete: Optional[str] = None
    count: Optional[str] = None


class MinmodHelper:
    @lru_cache(maxsize=1)
    @staticmethod
    def get_deposit_type_to_id():
        deposit_type_resp = httpx.get(
            f"{MINMOD_API}/deposit_types",
            verify=False,
            timeout=None,
        )
        deposit_type_resp.raise_for_status()

        dpt2id = {}
        for record in deposit_type_resp.json():
            assert record["name"] not in dpt2id
            dpt2id[record["name"]] = record["uri"][len(MNR_NS) :]
        return dpt2id

    @lru_cache(maxsize=1)
    @staticmethod
    def get_commodity_id2name():
        r = httpx.get(
            f"{MINMOD_API}/commodities",
            verify=False,
            timeout=None,
        )
        r.raise_for_status()

        id2name = {}
        for record in r.json():
            id = record["uri"][len(MNR_NS) :]
            name = record["name"]
            assert id not in id2name, (id, id2name)
            id2name[id] = name
        return id2name

    @lru_cache(maxsize=1)
    @staticmethod
    def get_unit_uri2name():
        r = httpx.get(
            f"{MINMOD_API}/units",
            verify=False,
            timeout=None,
        )
        r.raise_for_status()

        uri2name = {}
        for record in r.json():
            uri = record["uri"]
            name = record["name"]
            assert uri not in uri2name
            uri2name[uri] = name
        return uri2name


class CDRHelper:
    DedupSites = Endpoint(
        item="dedup-site",
        collection="dedup-sites",
        bulk_upload="dedup-sites",
        bulk_delete="dedup-site/bulk/sites",
        count="dedup-site/stats/count",
    )
    DepositType = Endpoint(item="deposit-type", collection="deposit-types")

    N_PARALLEL_JOBS = 16

    @staticmethod
    def upload_collection(endpoint: Endpoint, collection: list[dict]):
        with timer.Timer().watch_and_report(
            "Upload the whole collection to endpoint " + endpoint.collection
        ):
            if endpoint.bulk_upload is not None:
                r = httpx.post(
                    f"{CDR_API}/minerals/{endpoint.bulk_upload}",
                    json=collection,
                    headers=cdr_headers,
                    timeout=None,
                )
                r.raise_for_status()
            else:
                it = get_parallel(
                    n_jobs=CDRHelper.N_PARALLEL_JOBS, return_as="generator_unordered"
                )(delayed(CDRHelper.create)(endpoint, item) for item in collection)
                for _ in tqdm(it, total=len(collection), desc="uploading records"):
                    pass

    @staticmethod
    def truncate(endpoint: Endpoint):
        with timer.Timer().watch_and_report(
            "Truncating endpoint " + endpoint.collection
        ):
            if endpoint.bulk_upload is not None:
                r = httpx.delete(
                    f"{CDR_API}/minerals/{endpoint.bulk_delete}",
                    params={"system": MINMOD_SYSTEM},
                    headers=cdr_headers,
                    timeout=None,
                )
                assert r.status_code == 204 or r.status_code == 404, r.text
            else:
                r = httpx.get(
                    f"{CDR_API}/minerals/{endpoint.collection}",
                    headers=cdr_headers,
                    timeout=None,
                )
                r.raise_for_status()
                records = r.json()
                it = get_parallel(
                    n_jobs=CDRHelper.N_PARALLEL_JOBS, return_as="generator_unordered"
                )(
                    delayed(CDRHelper.delete_by_id)(endpoint, item["id"])
                    for item in records
                )
                for _ in tqdm(it, total=len(records), desc="deleting records"):
                    pass

            # double check the results
            if endpoint.count is None:
                r = httpx.get(
                    f"{CDR_API}/minerals/{endpoint.collection}",
                    headers=cdr_headers,
                    timeout=None,
                )
                r.raise_for_status()
                assert len(r.json()) == 0
            else:
                r = httpx.get(
                    f"{CDR_API}/minerals/{endpoint.count}",
                    headers=cdr_headers,
                    timeout=None,
                )
                r.raise_for_status()
                assert int(r.text.strip()) == 0

    @staticmethod
    def delete_by_id(endpoint: Endpoint, id: str):
        r = httpx.delete(
            f"{CDR_API}/minerals/{endpoint.item}/{id}",
            headers=cdr_headers,
            timeout=None,
        )
        assert r.status_code == 404 or r.status_code == 204, r.text

    @staticmethod
    def create(endpoint: Endpoint, item: dict):
        r = httpx.post(
            f"{CDR_API}/minerals/{endpoint.item}",
            json=item,
            headers=cdr_headers,
            timeout=None,
        )
        r.raise_for_status()


@lru_cache(maxsize=2)
def get_parallel(n_jobs: int = -1, return_as: str = "list"):
    return Parallel(n_jobs=n_jobs, return_as=return_as)



from __future__ import annotations

import sqlite3
from typing import Annotated, Optional

from fastapi import Depends
from minmodkg.api.models.user import User, UserBase, UserCreate, UserPublic, UserUpdate
from minmodkg.config import DBFILE
from sqlmodel import Session, SQLModel, create_engine

dbconn = f"sqlite:///{DBFILE}"
connect_args = {"check_same_thread": False}
engine = create_engine(dbconn, connect_args=connect_args)


def create_db_and_tables():
    SQLModel.metadata.create_all(engine)


def get_session():
    with Session(engine) as session:
        yield session


SessionDep = Annotated[Session, Depends(get_session)]

__all__ = [
    "User",
    "UserBase",
    "UserCreate",
    "UserPublic",
    "UserUpdate",
    "SessionDep",
    "create_db_and_tables",
    "get_session",
]



from __future__ import annotations

from datetime import datetime, timezone
from hashlib import sha256
from typing import Optional

from pydantic import BaseModel, Field


class CandidateExtractedEntity(BaseModel):
    source: str
    confidence: float
    observed_name: Optional[str] = None
    normalized_uri: Optional[str] = None


class LocationInfo(BaseModel):
    country: Optional[list[CandidateExtractedEntity]] = None
    state_or_province: Optional[list[CandidateExtractedEntity]] = None
    crs: Optional[CandidateExtractedEntity] = None
    location: Optional[str] = None


class Document(BaseModel):
    doi: Optional[str]
    uri: str
    title: Optional[str]


class BoundingBox(BaseModel):
    x_max: float
    x_min: float
    y_max: float
    y_min: float

    def to_enc_str(self):
        return f"BB:{self.x_max:.3f}_{self.x_min:.3f}_{self.y_max:.3f}_{self.y_min:.3f}"


class PageInfo(BaseModel):
    bounding_box: Optional[BoundingBox] = None
    page: int

    def to_enc_str(self):
        if self.bounding_box is None:
            return str(self.page)
        return f"PI:{self.page}|{self.bounding_box.to_enc_str()}"


class Reference(BaseModel):
    document: Document
    page_info: list[PageInfo]
    comment: Optional[str]
    property: Optional[str]


class MineralSiteBase(BaseModel):
    source_id: str
    record_id: str
    name: str
    location_info: Optional[LocationInfo] = None
    deposit_type_candidate: list[CandidateExtractedEntity] = Field(default_factory=list)
    reference: list[Reference] = Field(default_factory=list)


class MineralSite(MineralSiteBase):
    modified_at: str = Field(
        default_factory=lambda: datetime.now(timezone.utc).strftime(
            "%Y-%m-%dT%H:%M:%SZ"
        )
    )
    created_by: str


class MineralSiteCreate(MineralSiteBase):
    same_as: list[str] = Field(default_factory=list)


class MineralSiteUpdate(MineralSiteBase):
    same_as: list[str] = Field(default_factory=list)


# class MineralSiteUpdate(BaseModel):
#     source: Optional[str] = None
#     source_prop: Optional[str] = None
#     prop: str
#     object: str


# def get_site_uri(source_id: str, record_id)



from __future__ import annotations

import os
from pathlib import Path

import serde.yaml

if "CFG_FILE" not in os.environ:
    CFG_FILE = Path(__file__).parent.parent / "config.yml"
else:
    CFG_FILE = Path(os.environ["CFG_FILE"])
assert CFG_FILE.exists(), f"Config file {CFG_FILE} does not exist"
cfg = serde.yaml.deser(CFG_FILE)

# for sparql/ontology
SPARQL_ENDPOINT = cfg["triplestore"]["query"]
SPARQL_UPDATE_ENDPOINT = cfg["triplestore"]["update"]
MNR_NS = cfg["mnr_ns"]
MNO_NS = cfg["mno_ns"]
assert MNR_NS.endswith("/") or MNR_NS.endswith(
    "#"
), f"MNR_NS {MNR_NS} must end with / or #"
assert MNO_NS.endswith("/") or MNO_NS.endswith(
    "#"
), f"MNO_NS {MNO_NS} must end with / or #"

API_PREFIX = cfg["api_prefix"]
LOD_PREFIX = cfg["lod_prefix"]

while API_PREFIX.endswith("/"):
    API_PREFIX = API_PREFIX[:-1]

while LOD_PREFIX.endswith("/"):
    LOD_PREFIX = LOD_PREFIX[:-1]

# for databases
DBFILE = Path(cfg["dbfile"])
DBFILE.parent.mkdir(parents=True, exist_ok=True)

# for login/security
SECRET_KEY = cfg["secret_key"]
JWT_ALGORITHM = "HS256"
JWT_ACCESS_TOKEN_EXPIRE_MINUTES = 60 * 24 * 3  # 3 days



from __future__ import annotations

from functools import lru_cache

from fastapi import APIRouter
from minmodkg.api.dependencies import SPARQL_ENDPOINT, get_snapshot_id
from minmodkg.misc import sparql_query
from minmodkg.transformations import make_site_uri

router = APIRouter(tags=["units"])


@router.get("/units")
def units():
    return get_units(get_snapshot_id())


@lru_cache(maxsize=1)
def get_units(snapshot_id: str, endpoint: str = SPARQL_ENDPOINT):
    query = """
    SELECT ?uri ?name
    WHERE {
        ?uri a :Unit ;
            rdfs:label ?name .
    }
    """
    qres = sparql_query(query, endpoint)
    return qres



from __future__ import annotations

from datetime import datetime, timedelta, timezone
from hashlib import sha256
from typing import Annotated
from uuid import uuid4

import jwt
from fastapi import APIRouter, HTTPException, Query, Response, status
from jwt.exceptions import InvalidTokenError
from minmodkg.api.dependencies import CurrentUserDep, TokenDep
from minmodkg.api.models.db import SessionDep
from minmodkg.api.models.user import User, UserCreate, UserPublic
from minmodkg.config import JWT_ACCESS_TOKEN_EXPIRE_MINUTES, JWT_ALGORITHM, SECRET_KEY
from passlib.context import CryptContext
from sqlmodel import select

router = APIRouter(tags=["admin"])


@router.get("/users", response_model=list[UserPublic])
def get_users(
    session: SessionDep,
    user: CurrentUserDep,
    offset: int = 0,
    limit: Annotated[int, Query(le=100)] = 100,
):
    """Get all users"""
    users = session.exec(select(User).offset(offset).limit(limit)).all()
    return users


@router.post("/users", response_model=UserPublic)
def create_user(user: UserCreate, session: SessionDep, current_user: CurrentUserDep):
    """Create a new user"""
    if current_user.scope != "admin":
        raise HTTPException(403, "You do not have permission to create a new user")
    return create_user_priv(user, session)


def create_user_priv(user: UserCreate, session: SessionDep):
    exist_user = session.get(User, user.username)
    if exist_user is not None:
        raise HTTPException(400, "Username already exists")

    dbuser = User.model_validate(user)
    dbuser.encrypt_password()

    session.add(dbuser)
    session.commit()
    session.refresh(dbuser)
    return dbuser



from __future__ import annotations

from contextlib import contextmanager
from typing import Annotated

import typer
from minmodkg.api.internal.admin import create_user_priv
from minmodkg.api.models.db import create_db_and_tables, get_session
from minmodkg.api.models.user import UserCreate

app = typer.Typer(pretty_exceptions_short=True, pretty_exceptions_enable=False)


@app.command()
def user(
    username: Annotated[str, typer.Option("-u", help="Username")],
    name: Annotated[str, typer.Option("-n", help="Name")],
    email: Annotated[str, typer.Option("-e", help="Email")],
    password: Annotated[
        str, typer.Option(prompt=True, confirmation_prompt=True, hide_input=True)
    ],
):
    create_db_and_tables()

    with contextmanager(get_session)() as session:
        create_user_priv(
            UserCreate(username=username, name=name, email=email, password=password),
            session,
        )


if __name__ == "__main__":
    app()



from __future__ import annotations

from contextlib import contextmanager
from datetime import datetime, timedelta, timezone
from functools import lru_cache
from hashlib import sha256
from importlib.metadata import version
from pathlib import Path
from time import time
from typing import Callable, Optional
from uuid import uuid4

from drepr.main import convert
from drepr.models.resource import ResourceDataObject
from fastapi import APIRouter, HTTPException, status
from libactor.cache import BackendFactory, cache
from loguru import logger
from minmodkg.api.dependencies import CurrentUserDep
from minmodkg.api.models.mineral_site import (
    MineralSite,
    MineralSiteCreate,
    MineralSiteUpdate,
)
from minmodkg.config import MNO_NS, MNR_NS, SPARQL_ENDPOINT, SPARQL_UPDATE_ENDPOINT
from minmodkg.misc import (
    Transaction,
    TransactionError,
    Triples,
    has_uri,
    sparql,
    sparql_construct,
    sparql_delete_insert,
    sparql_insert,
    sparql_query,
)
from minmodkg.transformations import make_site_uri
from rdflib import Graph, URIRef

router = APIRouter(tags=["mineral_sites"])


@router.get("/mineral-sites/make-id")
def get_site_uri(source_id: str, record_id: str):
    return make_site_uri(source_id, record_id)


@router.post("/mineral-sites")
def create_site(site: MineralSiteCreate, user: CurrentUserDep):
    """Create a mineral site."""
    # To safely update the data, we need to do it in a transaction
    # There are two places where we update the data:
    #   1. The mineral site itself
    #   2. The dedup mineral site. However, the dedup mineral site can only be updated
    #      by reading the data first. This is impossible to do via Restful API as we
    #      cannot explicitly control the transaction. We have to achieve this by implementing
    #      a custom lock mechanism. We will revisit this later.
    uri = make_site_uri(site.source_id, site.record_id)

    if has_uri(uri):
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="The site already exists.",
        )

    # convert the site to TTL
    full_site = MineralSite(
        **site.model_dump(exclude={"same_as", "modified_at"}),
        modified_at=datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ"),
        created_by=MNR_NS + f"users/{user.username}",
    )
    g = get_mineral_site_model()(
        ResourceDataObject([full_site.model_dump(exclude_none=True)])
    )
    reluri = uri.replace(MNR_NS, "mnr:")

    # send the query
    resp = sparql_insert(
        [
            g,
            Triples(
                [
                    (reluri, "owl:sameAs", same_as.replace(MNR_NS, "mnr:"))
                    for same_as in site.same_as
                ]
            ),
        ],
    )
    if resp.status_code != 200:
        logger.error(
            "Failed to create site:\n\t- User: {}\n\t- Input: {}\n\t- Response Code: {}\n\t- Response Message{}",
            user.username,
            site.model_dump_json(),
            resp.status_code,
            resp.text,
        )
        raise HTTPException(status_code=500, detail="Failed to create the site.")

    return {"status": "success", "uri": uri}


@router.post("/mineral-sites/{site_id}")
def update_site(site_id: str, site: MineralSiteUpdate, user: CurrentUserDep):
    uri = MNR_NS + site_id
    if not has_uri(uri):
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="The site does not exist.",
        )

    if user.is_system():
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="Automated system is not allowed to update the site.",
        )

    # transform the site to TTL
    full_site = MineralSite(
        **site.model_dump(exclude={"same_as", "modified_at"}),
        modified_at=datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ"),
        created_by=MNR_NS + f"users/{user.username}",
    )
    ng = get_mineral_site_model()(
        ResourceDataObject([full_site.model_dump(exclude_none=True)])
    )
    # the site must have no blank nodes as we want a fast way to compute the delta.

    # start the transaction, we can only have one update at a time
    with Transaction([uri]).transaction():
        og = get_site(uri)
        del_triples, add_triples = get_site_changes(og, ng)
        sparql_delete_insert(del_triples, add_triples)

    return {"status": "success", "uri": uri}


def get_site(site_uri: str, endpoint: str = SPARQL_ENDPOINT) -> Graph:
    query = (
        """
CONSTRUCT {
    ?s ?p ?o .
    ?u ?e ?v .
}
WHERE {
    ?s ?p ?o .
    OPTIONAL {
        ?s (!(owl:sameAs|rdf:type))+ ?u .
        ?u ?e ?v .
        # FILTER (?e NOT IN (owl:sameAs))
    }
    VALUES ?s { <%s> }
}
"""
        % site_uri
    )
    return sparql_construct(query, endpoint)


def get_site_changes(current_site: Graph, new_site: Graph) -> tuple[Triples, Triples]:
    current_triples = set(current_site)
    new_triples = set(new_site)

    lock_pred = URIRef(MNO_NS + "lock")
    del_triples = Triples(
        triples=[
            (s.n3(), p.n3(), o.n3())
            for s, p, o in current_triples.difference(new_triples)
            if p != lock_pred
        ]
    )
    add_triples = Triples(
        triples=[
            (s.n3(), p.n3(), o.n3())
            for s, p, o in new_triples.difference(current_triples)
            if p != lock_pred
        ]
    )
    return del_triples, add_triples


@lru_cache()
def get_mineral_site_model() -> Callable[[ResourceDataObject], Graph]:
    pkg_dir = Path(__file__).parent.parent.parent
    drepr_version = version("drepr-v2").strip()

    @cache(
        backend=BackendFactory.func.sqlite.pickle(
            dbdir=pkg_dir / "extractors",
            mem_persist=True,
        ),
        cache_ser_args={
            "repr_file": lambda x: f"drepr::{drepr_version}::" + file_ident(x)
        },
    )
    def make_program(repr_file: Path, prog_file: Path):
        convert(
            repr=repr_file,
            resources={},
            progfile=prog_file,
        )

    # fix me! this is problematic when the prog_file is deleted but the cache is not cleared
    make_program(
        repr_file=pkg_dir.parent / "extractors/mineral_site.yml",
        prog_file=pkg_dir / "extractors/mineral_site.py",
    )
    from minmodkg.extractors.mineral_site import main  # type: ignore

    def map(resource: ResourceDataObject) -> Graph:
        ttl_data = main(resource)
        g = Graph()
        g.parse(data=ttl_data, format="turtle")
        return g

    return map


def file_ident(file: str | Path):
    file = Path(file).resolve()
    filehash = sha256(file.read_bytes()).hexdigest()
    return f"{file}::{filehash}"



from __future__ import annotations

from functools import lru_cache

from fastapi import APIRouter
from minmodkg.api.dependencies import SPARQL_ENDPOINT, get_snapshot_id
from minmodkg.misc import sparql_query

router = APIRouter(tags=["statistics"])


@router.get("/documents/count")
def document_count():
    return get_document_count(get_snapshot_id())


@router.get("/mineral-inventories/count")
def inventory_count():
    return get_inventory_count(get_snapshot_id())


@router.get("/mineral-sites/count")
def mineralsites_count():
    return get_mineralsites_count(get_snapshot_id())


@router.get("/mineral-inventories/count-by-commodity")
def inventory_by_commodity():
    return get_inventory_by_commodity(get_snapshot_id())


@router.get("/mineral-sites/count-by-commodity")
def mineralsites_by_commodity():
    return get_mineralsites_by_commodity(get_snapshot_id())


@router.get("/documents/count-by-commodity")
def documents_by_commodity():
    return get_documents_by_commodity(get_snapshot_id())


@lru_cache(maxsize=1)
def get_document_count(snapshot_id: str, endpoint: str = SPARQL_ENDPOINT):
    query = """
        SELECT (COUNT(?doc) AS ?total)
        WHERE {
            ?doc a :Document .
        }
    """
    qres = sparql_query(query, endpoint)
    return qres[0]


@lru_cache(maxsize=1)
def get_inventory_count(snapshot_id: str, endpoint: str = SPARQL_ENDPOINT):
    query = """
        SELECT (COUNT(?mi) AS ?total)
        WHERE {
            ?mi a :MineralInventory .
        }
    """
    qres = sparql_query(query, endpoint)
    return qres[0]


@lru_cache(maxsize=1)
def get_mineralsites_count(snapshot_id: str, endpoint: str = SPARQL_ENDPOINT):
    query = """
       SELECT (COUNT(?ms) AS ?total)
        WHERE {
            ?ms a :MineralSite .
        }
    """
    qres = sparql_query(query, endpoint)
    return qres[0]


@lru_cache(maxsize=1)
def get_inventory_by_commodity(snapshot_id: str, endpoint: str = SPARQL_ENDPOINT):
    query = """
        SELECT  ?commodity_uri ?commodity_label ?total
        WHERE {
            {
                SELECT ?commodity_uri (COUNT(DISTINCT ?mi) AS ?total)
                WHERE {
                    ?mi a :MineralInventory .
                    ?mi :commodity/:normalized_uri ?commodity_uri .
                }
                GROUP BY ?commodity_uri
            }
            ?commodity_uri rdfs:label ?commodity_label .
        }
    """
    qres = sparql_query(query, endpoint)
    return qres


@lru_cache(maxsize=1)
def get_mineralsites_by_commodity(snapshot_id: str, endpoint: str = SPARQL_ENDPOINT):
    query = """
        SELECT  ?commodity_uri ?commodity_label ?total
        WHERE {
            {
                SELECT ?commodity_uri (COUNT(DISTINCT ?ms) AS ?total)
                WHERE {
                    ?ms a :MineralSite .
                    ?ms :mineral_inventory ?mi .
                    ?mi :commodity/:normalized_uri ?commodity_uri .
                }
                GROUP BY ?commodity_uri
            }
            ?commodity_uri rdfs:label ?commodity_label .
        }
        
    """
    qres = sparql_query(query, endpoint)
    return qres


@lru_cache(maxsize=1)
def get_documents_by_commodity(snapshot_id: str, endpoint: str = SPARQL_ENDPOINT):
    query = """

        SELECT  ?commodity_uri ?commodity_label ?total
        WHERE {
            {
                SELECT ?commodity_uri (COUNT(DISTINCT ?doc) AS ?total)
                WHERE {
                    ?mi :reference/:document ?doc . 
                    ?mi :commodity/:normalized_uri ?commodity_uri .
                }
                GROUP BY ?commodity_uri
            }
            ?commodity_uri rdfs:label ?commodity_label .
        }
    """
    qres = sparql_query(query, endpoint)
    return qres



from __future__ import annotations

import hashlib
from typing import Callable
from uuid import uuid4

from drepr.models.preprocessing import Context
from minmodkg.api.models.mineral_site import PageInfo
from minmodkg.config import MNO_NS, MNR_NS
from slugify import slugify


def make_site_ids(value: dict, namespace: str = MNR_NS):
    """Make all ids for a mineral site"""
    site_uri = make_site_uri(value["source_id"], value["record_id"], namespace)
    site_id = site_uri[len(namespace) :]

    # we create id for each reference as it is likely to be the same or re-ordered
    value["id"] = site_uri
    for ref in value["reference"]:
        make_reference_ids(ref, site_id, namespace)
    for inv in value.get("mineral_inventory", []):
        make_reference_ids(inv["reference"], site_id, namespace)

    # for rest, we create id based on its index
    if "location_info" in value:
        make_location_info_ids(value["location_info"], site_id, namespace)

    for deptype_i, deptype in enumerate(value.get("deposit_type_candidate", [])):
        deptype["id"] = f"{namespace}{site_id}__deptype__{deptype_i}"

    for inv_i, inv in enumerate(value.get("mineral_inventory", [])):
        make_inventory_ids(inv, inv_i, site_id, namespace)


def make_location_info_ids(loc_info: dict, site_id: str, namespace: str = MNR_NS):
    loc_info["id"] = namespace + site_id + "__location_info"
    for i, country in enumerate(loc_info.get("country", [])):
        country["id"] = namespace + site_id + f"__country__{i}"
    for i, state in enumerate(loc_info.get("state_or_province", [])):
        state["id"] = namespace + site_id + f"__state__{i}"
    if loc_info.get("crs") is not None:
        loc_info["crs"]["id"] = namespace + site_id + f"__crs"


def make_inventory_ids(
    inv: dict, inv_index: int, site_id: str, namespace: str = MNR_NS
):
    inv_id = f"{site_id}__inv__{inv_index}"
    inv["id"] = f"{namespace}{inv_id}"
    for cat_idx, cat in enumerate(inv.get("category", [])):
        cat["id"] = f"{namespace}{inv_id}__cat__{cat_idx}"
    for key in ["commodity", "material_form"]:
        if inv.get(key) is not None:
            inv[key]["id"] = f"{namespace}{inv_id}__{key}"
    for key in ["cutoff_grade", "grade", "ore"]:
        if inv.get(key) is not None:
            inv[key]["id"] = f"{namespace}{inv_id}__{key}"
            if inv[key].get("unit") is not None:
                inv[key]["unit"]["id"] = f"{namespace}{inv_id}__{key}__unit"
    make_reference_ids(inv["reference"], site_id, namespace)


def make_reference_ids(ref: dict, site_id: str, namespace: str = MNR_NS):
    ref["document"]["id"] = make_document_uri(ref["document"], site_id, namespace)

    docid = ref["document"]["id"]
    if docid.startswith(namespace):
        docid = docid[len(namespace) :]
    elif docid.startswith("https://"):
        docid = docid[8:]
    elif docid.startswith("http://"):
        docid = docid[7:]
    docid = slugify(docid)

    ref["id"] = make_reference_uri(ref, slugify(docid), namespace)
    for i, pageinfo in enumerate(ref.get("page_info", [])):
        pageinfo["id"] = f"{ref['id']}__pageinfo__{i}"


def make_site_uri(source_id: str, record_id: str, namespace: str = MNR_NS) -> str:
    if source_id.find("::") != -1:
        # we need to remove the category from the source_id
        category, source_id = source_id.split("::", 1)
        assert category in {
            "mining-report",
            "article",
            "database",
            "curated-mining-report",
            "curated-article",
            "curated-database",
        }

    if source_id.startswith("http://"):
        if source_id.startswith("http://"):
            source_id = source_id[7:]
            if source_id.endswith("/"):
                source_id = source_id[:-1]
    elif source_id.startswith("https://"):
        source_id = source_id[8:]
        if source_id.endswith("/"):
            source_id = source_id[:-1]

    source_id = slugify(source_id)

    if isinstance(record_id, int):
        record_id = str(record_id)
    else:
        record_id = slugify(record_id)

    path = shorten_id(f"{source_id}__{record_id}", 120)
    return f"{namespace}site__{path}"


def make_document_uri(doc: dict, site_id: str, namespace: str = MNR_NS):
    if "uri" in doc:
        return doc["uri"]
    if "doi" in doc:
        assert not (
            doc["doi"].startswith("https://")
            or doc["doi"].startswith("http://")
            or doc["doi"].startswith("doi:")
            or doc["doi"].startswith("/")
        )
        return "https://doi.org/" + doc["doi"]

    if "title" not in doc:
        raise ValueError("Document must have a URI, DOI, or at least a title")

    path = site_id + "__doc__" + shorten_id(slugify(doc["title"]), 120)
    return f"{namespace}{path}"


def make_reference_uri(ref: dict, doc_id: str, namespace: str = MNR_NS):
    # gen pageinfo id
    if len(ref.get("page_info", [])) > 0:
        pageinfo_id = hashlib.sha256(
            b"|".join(
                (
                    PageInfo.model_validate(page_info).to_enc_str().encode()
                    for page_info in ref["page_info"]
                )
            )
        ).hexdigest()[:16]
    else:
        pageinfo_id = ""

    if "property" in ref:
        property = ref["property"]
        assert property.startswith(MNO_NS)
        property = property[len(MNO_NS) :]
    else:
        property = ""

    constraintinfo = property + "_" + pageinfo_id
    if len(constraintinfo) == 1:
        return namespace + doc_id + "__ref"
    else:
        return namespace + doc_id + "__ref__" + shorten_id(slugify(constraintinfo), 120)


def get_uuid4(prefix: str = "B", namespace: str = MNR_NS):
    return f"{namespace}{prefix}_{str(uuid4()).replace('-', '')}"


def shorten_id(long_id: str, max_length: int = 120):
    if len(long_id) > max_length:
        return (
            long_id[:max_length]
            + "__"
            + hashlib.sha256(long_id.encode()).hexdigest()[:8]
        )
    return long_id



from __future__ import annotations

from datetime import datetime, timedelta, timezone
from hashlib import sha256
from typing import Annotated
from uuid import uuid4

import jwt
from fastapi import APIRouter, HTTPException, Query, Response, status
from jwt.exceptions import InvalidTokenError
from minmodkg.api.dependencies import CurrentUserDep, TokenDep
from minmodkg.api.models.db import SessionDep
from minmodkg.api.models.user import User, UserCreate, UserPublic
from minmodkg.config import JWT_ACCESS_TOKEN_EXPIRE_MINUTES, JWT_ALGORITHM, SECRET_KEY
from passlib.context import CryptContext
from sqlmodel import select

router = APIRouter(tags=["login"])


@router.post("/login")
async def login(
    session: SessionDep,
    response: Response,
    username: str,
    password: str,
):
    # authenticate user
    user = session.get(User, username)
    if user is None or not user.verify_password(password):
        # validate password
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Incorrect username or password",
        )

    expired_at = datetime.now(timezone.utc) + timedelta(
        minutes=JWT_ACCESS_TOKEN_EXPIRE_MINUTES
    )
    access_token = jwt.encode(
        {"sub": user.username, "exp": expired_at.timestamp()},
        SECRET_KEY,
        algorithm=JWT_ALGORITHM,
    )
    response.set_cookie(
        key="session",
        value=access_token,
    )
    return "Logged in"


@router.get("/whoami", response_model=UserPublic)
def whoami(user: CurrentUserDep):
    return user



from minmodkg.misc.exceptions import TransactionError, UnconvertibleUnitError
from minmodkg.misc.geo import merge_wkt, merge_wkts, reproject_wkt
from minmodkg.misc.prefix_index import LongestPrefixIndex
from minmodkg.misc.sparql import (
    Transaction,
    Triples,
    has_uri,
    sparql,
    sparql_construct,
    sparql_delete_insert,
    sparql_insert,
    sparql_query,
)
from minmodkg.misc.utils import V, assert_isinstance, batch, group_by_attr, group_by_key

__all__ = [
    "V",
    "TransactionError",
    "UnconvertibleUnitError",
    "assert_isinstance",
    "batch",
    "group_by_attr",
    "group_by_key",
    "LongestPrefixIndex",
    "merge_wkt",
    "merge_wkts",
    "reproject_wkt",
    "sparql_insert",
    "sparql_query",
    "sparql_construct",
    "sparql",
    "Triples",
    "Transaction",
    "has_uri",
    "sparql_delete_insert",
]



from __future__ import annotations

import hashlib
import importlib
import re
import subprocess
import sys
from dataclasses import dataclass
from functools import cached_property, lru_cache
from importlib.metadata import version
from pathlib import Path
from typing import Annotated, Callable, NotRequired, TypedDict

import orjson
import typer
from drepr.main import convert
from loguru import logger
from rdflib import RDF, SH, Graph
from tqdm.auto import tqdm

from statickg.helper import CacheProcess, import_func
from statickg.models.file_and_path import BaseType, RelPath
from statickg.models.prelude import ETLOutput, RelPath, Repository
from statickg.services.interface import BaseFileService


class FilenameValidatorServiceConstructArgs(TypedDict):
    pattern: str
    verbose: NotRequired[int]


class FilenameValidatorServiceInvokeArgs(TypedDict):
    input: RelPath | list[RelPath]


class FilenameValidatorService(BaseFileService[FilenameValidatorServiceInvokeArgs]):

    def forward(
        self,
        repo: Repository,
        args: FilenameValidatorServiceInvokeArgs,
        output: ETLOutput,
    ):
        verbose = self.args.get("verbose", 1)
        infiles = self.list_files(
            repo,
            args["input"],
            unique_filepath=False,
            optional=False,
            compute_missing_file_key=False,
        )

        regex = re.compile(self.args["pattern"])

        invalid_filenames = []
        for infile in tqdm(infiles, desc="Validate file names", disable=verbose != 1):
            if not regex.match(infile.path.stem):
                invalid_filenames.append(infile.relpath)

        if len(invalid_filenames) > 0:
            self.logger.error(
                "Invalid filenames: \n{}",
                "\n".join(f"\t{x}" for x in invalid_filenames),
            )
            raise Exception("Found invalid filenames")


class ContentValidator:

    @dataclass
    class DReprModel:
        key: str
        validate_model: Callable[[Path], None]
        transform_model: Callable[[Path], str]

    def __init__(self, workdir: Path, shacl_bin: Path, predefined_ent_dir: RelPath):
        self.workdir = workdir
        self.predefined_ent_dir = predefined_ent_dir
        self.shacl_bin = shacl_bin
        self.shape_file = Path(__file__).parent.parent / "schema/shapes.ttl"

        assert shacl_bin.exists(), f"SHACL binary not found at {shacl_bin}"

        # create additional directories
        (workdir / "mineral-sites").mkdir(parents=True, exist_ok=True)
        (workdir / "entities").mkdir(parents=True, exist_ok=True)

        self.extractor_dir = RelPath(
            basetype=BaseType.CFG_DIR,
            basepath=Path(__file__).parent.parent,
            relpath="extractors",
        )
        self.pkgdir = self.setup(workdir)
        self.drepr_version = version("drepr-v2").strip()
        self.cache = CacheProcess(workdir / "cache.db")

    def setup(self, workdir: Path):
        pkgname = "gen_validate_programs"
        pkgdir = workdir / pkgname

        try:
            m = importlib.import_module(pkgname)
            if Path(m.__path__[0]) != pkgdir:
                raise ValueError(
                    f"Existing a python package named {pkgname}, please uninstall it because it is reserved to store generated DREPR programs"
                )
        except ModuleNotFoundError:
            # we can use services as the name of the folder containing the services as it doesn't conflict with any existing
            # python packages
            pass

        pkgdir.mkdir(parents=True, exist_ok=True)
        (pkgdir / "__init__.py").touch(exist_ok=True)

        # add the package to the path
        if str(pkgdir.parent) not in sys.path:
            sys.path.insert(0, str(pkgdir.parent))

        return pkgdir

    def get_cache_key(self, filepath: RelPath):
        return f"validator:{self.drepr_version}:{hashlib.sha256(filepath.get_path().read_bytes()).hexdigest()}"

    @cached_property
    def site_model(self):
        return self._get_drepr_model("mineral_site")

    @cached_property
    def predefined_entity_models(self):
        return {
            key: self._get_drepr_model(key)
            for key in [
                "category",
                "commodity",
                "deposit_type",
                "unit",
                "country",
                "state_or_province",
                "epsg",
                "material_form",
            ]
        }

    @cached_property
    def _predefined_entity_graph(self):
        data = []
        for file in self.predefined_ent_dir.iterdir():
            if file.suffix != ".json" and file.suffix != ".csv":
                continue

            outfile = self.workdir / "entities" / (file.stem + ".ttl")
            self._exec_drepr_model(
                self.predefined_entity_models[file.stem], file, outfile
            )
            data.append(outfile.read_text())
        return "\n".join(data)

    def _get_drepr_model(self, name: str):
        model_file = self.extractor_dir / f"{name}.yml"
        model_filepath = model_file.get_path()

        model_outfile = self.pkgdir / f"{model_filepath.stem}.py"
        model_key = self.get_cache_key(model_file)

        with self.cache.auto(
            filepath=model_file.get_ident(),
            key=model_key,
        ) as notfound:
            if notfound:
                try:
                    convert(
                        repr=model_filepath,
                        resources={},
                        progfile=model_outfile,
                    )
                except:
                    logger.error(
                        "Error when generating program {}", model_file.get_ident()
                    )
                    raise
                logger.info("generate program {}", model_file.get_ident())

        return ContentValidator.DReprModel(
            key=model_key,
            validate_model=lambda x: None,
            transform_model=import_func(
                f"{model_outfile.parent.name}.{model_outfile.stem}.main"
            ),
        )

    def _exec_drepr_model(
        self, model: ContentValidator.DReprModel, infile: RelPath, outfile: Path
    ):
        with self.cache.auto(
            filepath=infile.get_ident(),
            key=self.get_cache_key(infile) + ":model:" + model.key,
            outfile=outfile,
        ) as notfound:
            if notfound:
                try:
                    output = model.transform_model(infile.get_path())
                except:
                    logger.error("Encounter error when converting data to TTL")
                    raise
                outfile.write_text(output)

    def validate_mineral_site(self, site_file: RelPath):
        logger.info("Validating file {}", site_file)
        outfile = self.workdir / "mineral-sites" / (site_file.stem + ".ttl")

        # ------- schema check - pass 1 -------
        logger.info("Check if the data is valid with D-REPR (part 1)...")

        # ------- convert the site file to ttl -------
        logger.info("Check if the data can be converted to TTL...")
        self._exec_drepr_model(self.site_model, site_file, outfile)
        logger.info("Check if the data can be converted to TTL... Success!")

        # ------- validate again using SHACL to ensure all references are correct -------
        logger.info("Check if the data is valid with SHACL (part 2)...")
        # prepare a graph
        full_graph = self.workdir / "mineral-sites" / (site_file.stem + ".full.ttl")
        with open(full_graph, "w") as f:
            f.write(self._predefined_entity_graph)
            f.write("\n")
            f.write(outfile.read_text())

        process = subprocess.run(
            [
                "bash",
                str(self.shacl_bin),
                "-datafile",
                str(full_graph),
                "-shapesfile",
                str(self.shape_file),
            ],
            stdout=subprocess.PIPE,
        )
        shacl_output = process.stdout.decode("utf-8")

        # removing warning
        shacl_output_lines = shacl_output.splitlines()
        for i, line in enumerate(shacl_output_lines):
            if not line.startswith("@prefix"):
                assert re.match(r"\d*", line) is not None
                assert "WARN" in line
            else:
                shacl_output = "\n".join(shacl_output_lines[i:])
                break

        if process.returncode != 0:
            print("=" * 80, "\n", "-" * 10, "SHACL output", "-" * 10, "\n")
            print(shacl_output)
            print("=" * 80)
            raise Exception("SHACL validation failed")
        else:
            g = Graph()
            try:
                g.parse(data=shacl_output, format="turtle")
            except:
                print("=" * 80, "\n", "-" * 10, "SHACL output", "-" * 10, "\n")
                print(shacl_output)
                print("=" * 80)
                raise Exception("SHACL validation failed")

            for subj in g.subjects(RDF.type, SH.ValidationReport):
                (conforms,) = list(g.objects(subj, SH.conforms))
                if not conforms:
                    print("=" * 80, "\n", "-" * 10, "SHACL output", "-" * 10, "\n")
                    print(shacl_output)
                    print("=" * 80)
                    raise Exception("SHACL validation failed")

        logger.info("Check if the data is valid with SHACL (part 2)... Success!")


if __name__ == "__main__":
    app = typer.Typer(pretty_exceptions_short=True, pretty_exceptions_enable=False)

    @app.command()
    def validate(
        datadir: Annotated[
            Path, typer.Argument(help="A directory containing the data Git repository")
        ],
        workdir: Annotated[
            Path,
            typer.Argument(help="A directory for storing intermediate ETL results"),
        ],
        shacl_bin: Annotated[
            Path,
            typer.Argument(
                help="A path to the SHACL binary",
            ),
        ],
        changed_files_lst: Annotated[
            Path,
            typer.Argument(
                help="A list of changed files (relative path) in the data repository",
            ),
        ],
    ):
        validator = ContentValidator(
            workdir,
            shacl_bin,
            RelPath(
                basetype=BaseType.DATA_DIR, basepath=datadir, relpath="data/entities"
            ),
        )

        if changed_files_lst.name.endswith(".json"):
            changed_files = orjson.loads(changed_files_lst.read_bytes())
        else:
            assert changed_files_lst.name.endswith(".txt")
            changed_files = changed_files_lst.read_text().splitlines()

        for relpath in changed_files:
            if relpath.startswith("data/entities/"):
                # validate the entities
                raise NotImplementedError()
            elif any(
                relpath.startswith(dir)
                for dir in [
                    "data/inferlink/extractions/",
                    "data/umn/",
                    "data/sri/",
                    "data/usc/",
                    "test-mineral-sites/",
                ]
            ) and relpath.endswith(".json"):
                print("::group::Validate file", relpath)
                validator.validate_mineral_site(
                    RelPath(
                        basetype=BaseType.DATA_DIR, basepath=datadir, relpath=relpath
                    )
                )
                print("::endgroup::")
            else:
                print("::group::Not validate file", relpath)
                print("::endgroup::")

    app()



from __future__ import annotations

from pathlib import Path
from typing import Iterable, Mapping, NotRequired, TypedDict

import networkx as nx
import serde.csv
import serde.json
from joblib import Parallel, delayed
from libactor.cache import cache
from tqdm import tqdm

from statickg.helper import FileSqliteBackend, logger_helper
from statickg.models.etl import ETLOutput
from statickg.models.file_and_path import (
    FormatOutputPath,
    FormatOutputPathModel,
    InputFile,
    RelPath,
)
from statickg.models.repository import Repository
from statickg.services.interface import BaseFileService, BaseService

"""
Create a dedup group sites that are the same as each other.

This service works in two steps:

1. For each file, it produces a mapping from each site to a local dedup group.
2. Then, we map each local dedup to a final dedup site. Hopefully this graph is much smaller and much sparse.
3. Then, we generate same-as and dedup group.

"""


class SameAsServiceConstructArgs(TypedDict):
    verbose: NotRequired[int]
    parallel: NotRequired[bool]


class SameAsServiceInvokeArgs(TypedDict):
    input: RelPath | list[RelPath]
    output: RelPath | FormatOutputPath
    optional: NotRequired[bool]
    compute_missing_file_key: NotRequired[bool]


class SameAsService(BaseFileService[SameAsServiceInvokeArgs]):
    def __init__(
        self,
        name: str,
        workdir: Path,
        args: SameAsServiceConstructArgs,
        services: Mapping[str, BaseService],
    ):
        super().__init__(name, workdir, args, services)
        self.verbose = args.get("verbose", 1)
        self.parallel = args.get("parallel", True)
        self.parallel_executor = Parallel(n_jobs=-1, return_as="generator_unordered")

    def forward(
        self, repo: Repository, args: SameAsServiceInvokeArgs, tracker: ETLOutput
    ):
        infiles = self.list_files(
            repo,
            args["input"],
            unique_filepath=True,
            optional=args.get("optional", False),
            compute_missing_file_key=args.get("compute_missing_file_key", True),
        )
        outfmt = FormatOutputPathModel.init(args["output"])

        jobs = []
        for infile in infiles:
            outfile = outfmt.get_outfile(infile.path)
            outfile.parent.mkdir(parents=True, exist_ok=True)
            group_prefix = outfile.parent / outfile.stem
            jobs.append((group_prefix, infile, outfile))

        readable_ptns = self.get_readable_patterns(args["input"])

        if self.parallel:
            it: Iterable = self.parallel_executor(
                delayed(step1_exec)(self.workdir, group_prefix, infile, outfile)
                for group_prefix, infile, outfile in jobs
            )
        else:
            it: Iterable = (
                step1_exec(self.workdir, group_prefix, infile, outfile)
                for group_prefix, infile, outfile in jobs
            )

        outfiles = set()
        for outfile in tqdm(
            it, desc=f"Generating same-as for {readable_ptns}", disable=self.verbose < 1
        ):
            outfiles.add(outfile.relative_to(outfmt.outdir))

        self.remove_unknown_files(outfiles, outfmt.outdir)


def step1_exec(workdir: Path, prefix: str, infile: InputFile, outfile: Path):
    return Step1Fn.get_instance(workdir).same_as_step1_exec(prefix, infile, outfile)


class Step1Fn:
    instances = {}

    def __init__(self, workdir: Path):
        self.workdir = workdir

    @staticmethod
    def get_instance(workdir: Path):
        if workdir not in Step1Fn.instances:
            Step1Fn.instances[workdir] = Step1Fn(workdir)
        return Step1Fn.instances[workdir]

    @cache(
        backend=FileSqliteBackend.factory(filename="same_as_step1_exec.v100.sqlite"),
        cache_ser_args={
            "infile": lambda x: x.get_ident(),
        },
    )
    def same_as_step1_exec(self, prefix: str, infile: InputFile, outfile: Path):
        edges = serde.csv.deser(infile.path)
        G = nx.from_edgelist(edges[1:])
        groups = nx.connected_components(G)

        mapping = {
            f"{prefix}:{gid}": list(group) for gid, group in enumerate(groups, start=1)
        }
        serde.json.ser(mapping, outfile)
        return outfile



from __future__ import annotations

from collections import defaultdict
from typing import Generic, Optional, Union

from minmodkg.misc.utils import V


class LongestPrefixIndex(Generic[V]):
    """Given a list of prefixes, we want to find the longest predefined prefixes of a given string."""

    def __init__(
        self,
        index: dict[str, LongestPrefixIndex | str],
        start: int,
        end: int,
    ) -> None:
        self.index = index
        self.start = start
        self.end = end

    @staticmethod
    def create(prefixes: list[str]):
        sorted_prefixes = sorted(prefixes, key=lambda x: len(x), reverse=True)
        if len(sorted_prefixes) == 0:
            raise Exception("No prefix provided")
        return LongestPrefixIndex._create(sorted_prefixes, 0)

    @staticmethod
    def _create(sorted_prefixes: list[str], start: int):
        shortest_ns = sorted_prefixes[-1]
        index = LongestPrefixIndex({}, start, len(shortest_ns))

        if index.start == index.end:
            index.index[""] = shortest_ns
            subindex = LongestPrefixIndex._create(sorted_prefixes[:-1], index.end)
            for key, node in subindex.index.items():
                assert key not in index.index
                index.index[key] = node
            index.end = subindex.end
            return index

        tmp = defaultdict(list)
        for i, prefix in enumerate(sorted_prefixes):
            key = prefix[index.start : index.end]
            tmp[key].append(i)

        for key, lst_prefix_idx in tmp.items():
            if len(lst_prefix_idx) == 1:
                index.index[key] = sorted_prefixes[lst_prefix_idx[0]]
            else:
                index.index[key] = LongestPrefixIndex._create(
                    [sorted_prefixes[i] for i in lst_prefix_idx], index.end
                )
        return index

    def get(self, s: str) -> Optional[str]:
        """Get prefix of a string. Return None if it is not found"""
        key = s[self.start : self.end]
        if key in self.index:
            prefix = self.index[key]
            if isinstance(prefix, LongestPrefixIndex):
                return prefix.get(s)
            return prefix if s.startswith(prefix) else None

        if "" in self.index:
            prefix = self.index[""]
            assert isinstance(prefix, str)
            return prefix if s.startswith(prefix) else None

        return None

    def __str__(self):
        """Readable version of the index"""
        stack: list[tuple[int, str, Union[str, LongestPrefixIndex]]] = list(
            reversed([(0, k, v) for k, v in self.index.items()])
        )
        out = []

        while len(stack) > 0:
            depth, key, value = stack.pop()
            indent = "    " * depth
            if isinstance(value, str):
                out.append(indent + "`" + key + "`: " + value + "\n")
            else:
                out.append(indent + "`" + key + "`:" + "\n")
                for k, v in value.index.items():
                    stack.append((depth + 1, k, v))

        return "".join(out)

    def to_dict(self):
        return {
            k: v.to_dict() if isinstance(v, LongestPrefixIndex) else v
            for k, v in self.index.items()
        }



from __future__ import annotations

from typing import Optional

import pandas as pd
import shapely.ops
from pyproj import Transformer
from shapely.geometry import GeometryCollection
from shapely.wkt import dumps, loads


def merge_wkt(series):
    geometries = []
    for wkt in series:
        if pd.notna(wkt) and isinstance(wkt, str):
            try:
                geometry = loads(wkt)
                geometries.append(geometry)
            except Exception as e:
                print(f"Warning: Error loading WKT: {e} for WKT: {wkt}, skipping entry")

    if len(geometries) == 1:
        # return the single geometry directly
        return dumps(geometries[0])
    elif len(geometries) > 1:
        # return a GEOMETRYCOLLECTION if there are multiple geometries
        return dumps(GeometryCollection(geometries))
    else:
        # return None if there are no valid geometries
        return None


def merge_wkts(
    lst: list[tuple[int, Optional[str], str]], min_rank: Optional[int] = None
) -> tuple[str, str]:
    """Merge a list of WKTS with potentially different CRS into a single WKT"""
    if min_rank is None:
        min_rank = max(x[0] for x in lst)
    norm_lst: list[tuple[str, str]] = [
        (crs or "EPSG:4326", wkt) for rank, crs, wkt in lst if rank >= min_rank
    ]
    all_crs = set(x[0] for x in norm_lst)
    if len(all_crs) == 0:
        norm_crs = ""
    elif len(all_crs) == 1:
        norm_crs = all_crs.pop()
    else:
        if "EPSG:4326" in all_crs:
            norm_crs = "EPSG:4326"
        else:
            norm_crs = all_crs.pop()

        # we convert everything to norm_crs
        norm_lst = [(crs, reproject_wkt(wkt, crs, norm_crs)) for crs, wkt in norm_lst]

    # all CRS are the same
    wkts = sorted({x[1] for x in norm_lst})
    if len(wkts) > 1:
        wkt = merge_wkt(wkts)
        if wkt is None:
            wkt = ""
    else:
        wkt = wkts[0]
    return norm_crs, wkt


def reproject_wkt(wkt: str, from_crs: str, to_crs: str) -> str:
    assert from_crs.startswith("EPSG:"), from_crs
    assert to_crs.startswith("EPSG:"), to_crs

    if from_crs == to_crs:
        return wkt

    transformer = Transformer.from_crs(
        int(from_crs[len("EPSG:") :]), int(to_crs[len("EPSG:") :])
    )

    return dumps(shapely.ops.transform(transformer.transform, loads(wkt)))



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timezone
from functools import lru_cache
from typing import Annotated, Optional

import jwt
from fastapi import Depends, HTTPException, status
from fastapi.security import APIKeyCookie
from minmodkg.api.models.db import SessionDep
from minmodkg.api.models.user import User
from minmodkg.config import JWT_ALGORITHM, MNR_NS, SECRET_KEY, SPARQL_ENDPOINT
from minmodkg.misc import LongestPrefixIndex, sparql_query

# for login/security
token_from_cookie = APIKeyCookie(name="session")
TokenDep = Annotated[str, Depends(token_from_cookie)]


async def get_current_user(session: SessionDep, token: TokenDep):
    credentials_exception = HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail="Could not validate credentials",
    )
    try:
        payload = jwt.decode(token, SECRET_KEY, algorithms=[JWT_ALGORITHM])
        username: str = payload.get("sub")
        if username is None:
            raise credentials_exception
        expired_at = datetime.fromtimestamp(payload.get("exp"), timezone.utc)
        if expired_at < datetime.now(timezone.utc):
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Credentials expired",
            )
    except jwt.InvalidTokenError:
        raise credentials_exception

    user = session.get(User, username)
    if user is None:
        raise credentials_exception
    return user


CurrentUserDep = Annotated[User, Depends(get_current_user)]


def get_snapshot_id(endpoint: str = SPARQL_ENDPOINT):
    query = "SELECT ?snapshot_id WHERE { mnr:kg dcterms:hasVersion ?snapshot_id }"
    qres = sparql_query(query, endpoint)
    return qres[0]["snapshot_id"]


def norm_commodity(commodity: str, endpoint: str = SPARQL_ENDPOINT) -> str:
    if commodity.startswith("http"):
        raise HTTPException(
            status_code=404,
            detail=f"Expect commodity to be either just an id (QXXX) or name. Get `{commodity}` instead",
        )
    if not is_minmod_id(commodity):
        uri = get_commodity_by_name(commodity, endpoint)
        if uri is None:
            raise HTTPException(
                status_code=404, detail=f"Commodity `{commodity}` not found"
            )
        commodity = uri
    return commodity


def is_minmod_id(text: str) -> bool:
    return text.startswith("Q") and text[1:].isdigit()


def get_commodity_by_name(name: str, endpoint: str = SPARQL_ENDPOINT) -> Optional[str]:
    query = (
        'SELECT ?uri WHERE { ?uri a :Commodity ; rdfs:label ?name . FILTER(LCASE(STR(?name)) = "%s") } LIMIT 1'
        % name.lower()
    )
    qres = sparql_query(query, endpoint)
    if len(qres) == 0:
        return None
    uri = qres[0]["uri"]
    assert uri.startswith(MNR_NS)
    uri = uri[len(MNR_NS) :]
    return uri


def rank_source(
    source_id: str, snapshot_id: str, endpoint: str = SPARQL_ENDPOINT
) -> int:
    """Get ranking of a source, higher is better"""
    default_score = 5
    score = get_source_scores(snapshot_id, endpoint).get_score(source_id)
    if score is None:
        print("Unknown source id:", source_id)
        return default_score
    return score


@dataclass
class SourceScore:
    source2score: dict[str, int]
    index: LongestPrefixIndex

    def get_score(self, source_id: str) -> Optional[int]:
        prefix = self.index.get(source_id)
        if prefix is not None:
            return self.source2score[prefix]
        return None


@lru_cache(maxsize=1)
def get_source_scores(snapshot_id: str, endpoint: str = SPARQL_ENDPOINT):
    query = """
    SELECT ?uri ?prefixed_id ?score
    WHERE {
        ?uri a :SourceScore ;
            :prefixed_id ?prefixed_id ;
            :score ?score
    }
    """
    qres = sparql_query(query, endpoint)
    source2score = {record["prefixed_id"]: record["score"] for record in qres}
    index = LongestPrefixIndex.create(list(source2score.keys()))
    return SourceScore(source2score, index)



"""
Copied from https://github.com/DARPA-CRITICALMAAS/cdr_schemas/blob/e05bd770b1d4a71eca200640b3d239e8c93e056f/cdr_schemas/mineral.py
"""

from typing import List, Optional, Union

from pydantic import BaseModel, Field


class DocumentReference(BaseModel):
    cdr_id: str
    page: Optional[int] = Field(default=None)
    x_min: Optional[float] = Field(default=None)
    x_max: Optional[float] = Field(default=None)
    y_min: Optional[float] = Field(default=None)
    y_max: Optional[float] = Field(default=None)


class EvidenceLayer(BaseModel):
    name: str = Field(default="")
    relevance_score: float


class MappableCriteria(BaseModel):
    criteria: str
    theoretical: str = Field(default="")
    potential_dataset: list[EvidenceLayer] = Field(
        default_factory=list, description="List of evidence layers"
    )
    supporting_references: list[DocumentReference]


class MineralSystem(BaseModel):
    deposit_type: list[str] = Field(default_factory=list)
    source: list[MappableCriteria] = Field(default_factory=list)
    pathway: list[MappableCriteria] = Field(default_factory=list)
    trap: list[MappableCriteria] = Field(
        default_factory=list, description="Mappable Criteria: trap"
    )
    preservation: list[MappableCriteria] = Field(
        default_factory=list, description="Mappable Criteria: Preservation"
    )
    energy: list[MappableCriteria] = Field(
        default_factory=list, description="Mappable Criteria: Energy"
    )
    outflow: list[MappableCriteria] = Field(
        default_factory=list, description="Mappable Criteria: outflow"
    )


class GeologyInfo(BaseModel):
    age: str = Field(default="", description="Age of the geologic unit or event")
    unit_name: str = Field(default="", description="Name of the geologic unit")
    description: str = Field(default="")
    lithology: List[str] = Field(default_factory=list, description="Lithology")
    process: List[str] = Field(default_factory=list, description="Process")
    environment: List[str] = Field(default_factory=list, description="environment")
    comments: str = Field(default="")


class DepositType(BaseModel):
    id: Optional[str] = Field(default=None, description="Deposit type id")
    name: str = Field(description="Deposit type name")
    environment: str = Field(description="Deposit type environment")
    group: str = Field(description="Deposit type group")


class DepositTypeCandidate(BaseModel):
    observed_name: str = Field(
        default="",
        description="Source dataset that the site info is retrieved from. e.g., MRDS",
    )

    deposit_type_id: Optional[str] = Field(default=None, description="Deposit type id")

    confidence: Optional[Union[float, int]] = Field(
        default=None, description="Score deposit type of an inventory item"
    )
    source: str = Field(
        description="Source of the classification (automated model version / SME / etc...)"
    )


class RecordReference(BaseModel):
    record_id: str = Field(default="", description="id in source")
    source: str = Field(default="", description="Source information")
    uri: str = Field(default="", description="uri of source")


class MineralInventoryCategory(BaseModel):
    category: str = Field(description="category name")
    confidence: Optional[Union[float, int]] = Field(
        default=None,
    )
    source: str = Field(
        description="Source of the classification (automated model version / SME / etc...)"
    )


class GeoLocationInfo(BaseModel):
    crs: str = Field(
        description="The Coordinate Reference System (CRS) of the location"
    )

    geom: str = Field(
        description="Type: Polygon or Point, value indicates the geolocation of the site"
    )


class Confidence(BaseModel):
    confidence: Optional[Union[float, int]] = Field(default=None)
    source: str = Field(
        description="Source of the classification (automated model version / SME / etc...)"
    )


class MineralInventory(BaseModel):
    contained_metal: Optional[float] = Field(
        default=None,
        description="The quantity of a contained metal in an inventory item",
    )
    commodity: str = Field(default="", description="The commodity of an inventory item")
    commodity_observed_name: str = Field(
        default="",
        description="The observed commodity in the source data (textual format)",
    )

    ore_unit: str = Field(
        default="",
        description="The unit in which ore quantity is measured, eg, metric tonnes",
    )
    ore_value: Optional[float] = Field(
        default=None, description="The value of ore quantity"
    )

    grade_unit: str = Field(
        default="", description="The unit in which grade is measured, eg, percent"
    )
    grade_value: Optional[float] = Field(default=None, description="The value of grade")

    cutoff_grade_unit: str = Field(
        default="", description="The unit in which grade is measured, eg, percent"
    )
    cutoff_grade_value: Optional[float] = Field(default=None)

    material_form: Optional[float] = Field(default=None)
    material_form_unit: str = Field(default="")
    material_form_conversion: Optional[float] = Field(default=None)

    confidence: Optional[Confidence] = Field(default=None)

    categories: List[MineralInventoryCategory] = Field(
        default_factory=list,
        description="""
            A list of categories
        """,
    )

    documents: List[DocumentReference] = Field(
        default_factory=list,
        description="""
            A list of document references
        """,
    )

    records: List[RecordReference] = Field(
        default_factory=list,
        description="""
            A list of records references from databases or other sources
        """,
    )

    date: str = Field(
        default="", description="When in the point of time mineral inventory valid"
    )
    zone: str = Field(
        default="",
        description="zone of mineral site where inventory item was discovered",
    )


class MineralSite(BaseModel):
    id: str = Field(description="Mineral Site Id")
    source_id: str = Field(
        default="",
        description="Source dataset that the site info is retrieved from. e.g., MRDS",
    )
    record_id: str = Field(
        default="",
        description="Unique ID of the record that the info is retrieved from e.g., 10022920",
    )
    name: str = Field(default="", description="Name of the mine, e.g., Tungsten Jim")

    site_rank: str = Field(default="")
    site_type: str = Field(default="")
    country: List[str] = Field(default_factory=list)
    province: List[str] = Field(default_factory=list)
    location: Optional[GeoLocationInfo] = Field(default=None)
    mineral_inventory: List[MineralInventory] = Field(
        default_factory=list,
        description="""
            A list of mineral inventories
        """,
    )

    deposit_type_candidate: List[DepositTypeCandidate] = Field(
        default_factory=list,
        description="""
            A list of deposit types candidates
        """,
    )

    validated: bool = Field(False, description="Validated by human")

    system: str = Field(
        ...,
        description="""
            The name of the system used.
        """,
    )
    system_version: str = Field(
        ...,
        description="""
            The version of the system used.
        """,
    )


class DedupSiteRecord(BaseModel):
    id: Optional[str] = Field(default=None, description="dedup site record record id")
    mineral_site_id: str = Field(description="Mineral Site Id")
    name: str = Field(default="", description="Mineral Site Name")
    country: str = Field(default="")
    province: str = Field(default="")
    site_rank: str = Field(default="")
    site_type: str = Field(default="")


class DedupSite(BaseModel):
    id: Optional[str] = Field(default=None, description="dedup site id")
    sites: List[DedupSiteRecord] = Field(
        default_factory=list, description="Mineral Sites"
    )

    commodity: str = Field(..., description="Commodity Name")
    contained_metal: Optional[float] = Field(default=None)
    contained_metal_units: str = Field(default="")
    tonnage: Optional[float] = Field(default=None)
    tonnage_units: str = Field(default="")
    grade: Optional[float] = Field(default=None)
    grade_units: str = Field(default="")

    crs: str = Field(
        default="", description="The Coordinate Reference System (CRS) of the location"
    )
    centroid: Optional[str] = Field(
        default="",
        description="Type: Point (center) of the geolocation of the site, merged if there is a collection",
    )

    geom: Optional[str] = Field(
        default="",
        description="Type: Polygon or Point, value indicates the geolocation of the site",
    )

    deposit_type_candidate: List[DepositTypeCandidate] = Field(
        default_factory=list,
        description="""
            A list of deposit types candidates
        """,
    )

    system: str = Field(..., description="The name of the system used.")
    system_version: str = Field(..., description="The version of the system used.")
    data_snapshot: str = Field(..., description="version of data")
    data_snapshot_date: str = Field(..., description="date of data")
